﻿Imports System.Collections

Public Class Heap
   Public Shared Function FromSelector(Of T)(ByVal selector As Func(Of T, IComparable), ByVal minHeap As Boolean) As Heap(Of T)
      Return New Heap(Of T)(selector, minHeap)
   End Function
   Public Shared Function FromSelector(Of T)(ByVal selector As Func(Of T, IComparable), ByVal minHeap As Boolean, ByVal limit As Integer) As LimitedHeap(Of T)
      Return New LimitedHeap(Of T)(selector, minHeap, limit)
   End Function
End Class

<Diagnostics.DebuggerDisplay("Count={Count}, MinHeap={MinHeap}")> _
Public Class Heap(Of T)
   'ein Heap ist ein sortierter, balancierter binärer Baum, im Array (hier: List(of T)) abgebildet.
   'Die beiden Children eines Parents liegen immer genau hinter dem doppelten Index der Parents.
   'Heap-Bedingung (bei MinHeap): beide Children eines Parents müssen größer sein.

   ''' <summary>return 1, 0, -1, depending on (IComparable) key should be parent, equal or child of an item</summary>
   Public ReadOnly CompareKey As Func(Of IComparable, T, Integer)
   Public ReadOnly Selector As Func(Of T, IComparable)
   Public ReadOnly MinHeap As Boolean = True
   Private _InnerList As New List(Of T)()

   Public Sub New(ByVal Selector As Func(Of T, IComparable), ByVal minHeap As Boolean)
      Me.Selector = Selector
      Me.MinHeap = minHeap
      CompareKey = If(minHeap, _
         Function(value As IComparable, item As T) value.CompareTo(Selector(item)), _
         Function(value As IComparable, item As T) Selector(item).CompareTo(value))
   End Sub

   Public Sub Clear()
      _InnerList.Clear()
   End Sub

   ''' <summary>clear Heap and return (MinHeap: descending) sorted List</summary>
   Public Function GetListSorted() As List(Of T)
      'in-place-sortierung:
      'die Schleife entnimmt vorne immer den TopNode, und schreibt ihn ans Ende.
      'Die Heap-Struktur verkürzt sich dabei jeweils um ein Element
      For cnt = _InnerList.Count - 1 To 1 Step -1
         _InnerList(cnt) = PopCore(cnt)
      Next
      GetListSorted = _InnerList
      _InnerList = New List(Of T)()
   End Function

   ''' <summary>Insert Item</summary>
   Public Overridable Sub Push(ByVal item As T)
      _InnerList.Add(Nothing)      ' leeren Platz am Ende schaffen
      'Durch "Aufrücken" wird die korrekte EinfügePosition für item bereitgestellt
      'läuft, ausgehend von letzter Position, von Parent zu Parent, bis einer kleiner ist als key
      Dim key = Selector(item)
      Dim iChild = _InnerList.Count - 1
      While iChild <> 0
         Dim iParent = (iChild - 1) >> 1
         If CompareKey(key, _InnerList(iParent)) >= 0 Then Exit While 'Parent ist kleiner
         _InnerList(iChild) = _InnerList(iParent)         'bisherigen Parent auf iChild schreiben
         iChild = iParent
      End While
      _InnerList(iChild) = item
   End Sub

   ''' <summary>remove Top-Item and return it</summary>
   Public Function Pop() As T
      Dim heapLength = _InnerList.Count - 1
      Pop = PopCore(heapLength)
      _InnerList.RemoveAt(heapLength)
   End Function

   ''' <summary>return Top-Item without remove it</summary>
   Public Function Peek() As T
      Return _InnerList(0)
   End Function

   Public ReadOnly Property Count() As Integer
      Get
         Return _InnerList.Count
      End Get
   End Property

   ''' <summary>(MinHeap): if bigger than TopNode - take item, and - in exchange - overwrite ByRef-Parameter with popped TopNode</summary>
   Public Function TryExchange(ByRef item As T) As Boolean
      If CompareKey(Selector(item), _InnerList(0)) <= 0 Then Return False 'ablehnen, wenn item kleiner
      Dim itm = Pop()
      Push(item)
      item = itm
      Return True
   End Function

   Private Function PopCore(ByVal uBound As Integer) As T
      'entnimmt das TopItem, ohne die Innerlist zu verkürzen
      PopCore = _InnerList(0)      'der erste, Rückgabewert - sein Platz wird frei
      If uBound = 0 Then Exit Function
      Dim homeless = _InnerList(uBound) 'der letzte wird "obdachlos"
      Dim key = Selector(homeless)
      'es wird der Platz gesucht, wo homeless einzusetzen ist
      'Suche mittels Aufrücken: immer das kleinere der beiden Kinder übernimmt den Parent-Platz
      'Bis ein Platz gefunden ist, an dem homeless noch kleiner ist, oder uBound erreicht
      'Da mit jedem Loop sich die Schrittweite verdoppelt, geht das sehr schnell
      Dim iParent = 0
      Dim iChild = 1
      Do
         ' Heap - binärer Baum im Array: die beiden Children eines Knotens
         ' liegen genau hinter der Verdopplung des Parent-Index'
         Dim child = _InnerList(iChild)
         Dim iChild2 = iChild + 1
         If iChild2 < uBound Then
            Dim child2 = _InnerList(iChild2)
            If CompareKey(Selector(child2), child) < 0 Then 'child2 < child
               iChild = iChild2
               child = child2
            End If
         End If
         If CompareKey(key, child) <= 0 Then Exit Do 'homeless < child
         _InnerList(iParent) = child
         iParent = iChild
         iChild = (iChild << 1) + 1
      Loop Until iChild >= uBound    'uBound erreicht oder überschritten
      _InnerList(iParent) = homeless
   End Function

End Class

''' <summary>note: limited MinHeap selects biggest values</summary>
Public Class LimitedHeap(Of T) : Inherits Heap(Of T)
   Public ReadOnly Limit As Integer
   Public Sub New(ByVal Selector As Func(Of T, IComparable), ByVal minHeap As Boolean, ByVal limit As Integer)
      MyBase.New(Selector, minHeap)
      Me.Limit = limit
   End Sub
   Public Overrides Sub Push(ByVal item As T)
      If Count < Limit Then
         MyBase.Push(item)
      Else
         TryExchange(item)
      End If
   End Sub
End Class